#include <cstring>

namespace emilua::gperf::detail { using std::size_t; using std::strcmp; }


#include <emilua/detail/landlock.hpp>
#include <emilua/core.hpp>

#include <sys/sysmacros.h>
#include <sys/capability.h>
#include <sys/mount.h>
#include <sys/prctl.h>

#include <linux/securebits.h>
#include <linux/seccomp.h>
#include <linux/filter.h>
#include <grp.h>

#include <boost/scope_exit.hpp>

#define EMILUA_DETAIL_INT_CONSTANT(X) \
    [](lua_State* L) -> int {         \
        lua_pushinteger(L, X);        \
        return 1;                     \
    }
namespace emilua {


static void check_last_error(lua_State* L, int last_error,
                             const char* perror_string)
{
    if (last_error != 0) {
        lua_getfield(L, LUA_GLOBALSINDEX, "errexit");
        if (lua_toboolean(L, -1)) {
            errno = last_error;
            perror(perror_string);

            // new state which is not memory-limited (LUA_HOOK_BUFFER_SIZE) just
            // to allocate the stacktrace string (we exit the process w/o
            // returning thread's control to the user program so it's fine to
            // skip explicit_bzero())
            lua_State* L2 = luaL_newstate();

            luaL_traceback(L2, L, nullptr, 1);
            fprintf(stderr, "%s\n", lua_tostring(L2, -1));

            std::exit(1);
        }
    }
};

#define CHECK_LAST_ERROR(L, last_error, str) \
    check_last_error(L, last_error, "<3>ipc_actor/init/" str)


} // namespace emilua
namespace emilua {


using detail::landlock_handled_access_fs;

} // namespace emilua
namespace emilua::gperf::detail {
/* C++ code produced by gperf version 3.1 */
/* Command-line: /usr/bin/gperf --language=C++ --enum --readonly-tables --struct-type --initializer-suffix=',{}' --class-name=gperf__src_linux_ipc_actor_sys_bindings_cpp_2  */
/* Computed positions: -k'' */

struct word_type_gperf__src_linux_ipc_actor_sys_bindings_cpp_2 { const char* name; 
                                    const char* (*action)(
                                        lua_State*, landlock_ruleset_attr&); };
/* maximum key range = 1, duplicates = 0 */

class gperf__src_linux_ipc_actor_sys_bindings_cpp_2
{
private:
  static inline unsigned int hash (const char *str, size_t len);
public:
  static const struct word_type_gperf__src_linux_ipc_actor_sys_bindings_cpp_2 *in_word_set (const char *str, size_t len);
};

inline /*ARGSUSED*/
unsigned int
gperf__src_linux_ipc_actor_sys_bindings_cpp_2::hash (const char *str, size_t len)
{
  return 0;
}

const struct word_type_gperf__src_linux_ipc_actor_sys_bindings_cpp_2 *
gperf__src_linux_ipc_actor_sys_bindings_cpp_2::in_word_set (const char *str, size_t len)
{
  enum
    {
      TOTAL_KEYWORDS = 1,
      MIN_WORD_LENGTH = 17,
      MAX_WORD_LENGTH = 17,
      MIN_HASH_VALUE = 0,
      MAX_HASH_VALUE = 0
    };

  static const struct word_type_gperf__src_linux_ipc_actor_sys_bindings_cpp_2 wordlist[] =
    {
      {"handled_access_fs", 
                                    [](lua_State* L, landlock_ruleset_attr& a) {
                                        if (lua_type(L, -1) != LUA_TTABLE) {
                                            return "invalid handled_access_fs";
                                        }

                                        auto r = landlock_handled_access_fs(L);
                                        if (r) {
                                            a.handled_access_fs = r.value();
                                            return (const char*)(NULL);
                                        } else {
                                            return r.error();
                                        }
                                    }}
    };

  if (len <= MAX_WORD_LENGTH && len >= MIN_WORD_LENGTH)
    {
      unsigned int key = hash (str, len);

      if (key <= MAX_HASH_VALUE)
        {
          const char *s = wordlist[key].name;

          if (*str == *s && !strcmp (str + 1, s + 1))
            return &wordlist[key];
        }
    }
  return 0;
}

} // namespace emilua::gperf::detail
namespace emilua::gperf::detail {
/* C++ code produced by gperf version 3.1 */
/* Command-line: /usr/bin/gperf --language=C++ --enum --readonly-tables --struct-type --initializer-suffix=',{}' --class-name=gperf__src_linux_ipc_actor_sys_bindings_cpp_3  */
/* Computed positions: -k'' */

struct word_type_gperf__src_linux_ipc_actor_sys_bindings_cpp_3 { const char* name; std::uint32_t action; };
/* maximum key range = 1, duplicates = 0 */

class gperf__src_linux_ipc_actor_sys_bindings_cpp_3
{
private:
  static inline unsigned int hash (const char *str, size_t len);
public:
  static const struct word_type_gperf__src_linux_ipc_actor_sys_bindings_cpp_3 *in_word_set (const char *str, size_t len);
};

inline /*ARGSUSED*/
unsigned int
gperf__src_linux_ipc_actor_sys_bindings_cpp_3::hash (const char *str, size_t len)
{
  return 0;
}

const struct word_type_gperf__src_linux_ipc_actor_sys_bindings_cpp_3 *
gperf__src_linux_ipc_actor_sys_bindings_cpp_3::in_word_set (const char *str, size_t len)
{
  enum
    {
      TOTAL_KEYWORDS = 1,
      MIN_WORD_LENGTH = 7,
      MAX_WORD_LENGTH = 7,
      MIN_HASH_VALUE = 0,
      MAX_HASH_VALUE = 0
    };

  static const struct word_type_gperf__src_linux_ipc_actor_sys_bindings_cpp_3 wordlist[] =
    {
      {"version",  1U << 0}
    };

  if (len <= MAX_WORD_LENGTH && len >= MIN_WORD_LENGTH)
    {
      unsigned int key = hash (str, len);

      if (key <= MAX_HASH_VALUE)
        {
          const char *s = wordlist[key].name;

          if (*str == *s && !strcmp (str + 1, s + 1))
            return &wordlist[key];
        }
    }
  return 0;
}

} // namespace emilua::gperf::detail
namespace emilua::gperf::detail {
/* C++ code produced by gperf version 3.1 */
/* Command-line: /usr/bin/gperf --language=C++ --enum --readonly-tables --struct-type --initializer-suffix=',{}' --class-name=gperf__src_linux_ipc_actor_sys_bindings_cpp_4  */
/* Computed positions: -k'' */

struct word_type_gperf__src_linux_ipc_actor_sys_bindings_cpp_4 { const char* name; 
                                const char* (*action)(
                                    lua_State*, landlock_path_beneath_attr&); };
/* maximum key range = 6, duplicates = 0 */

class gperf__src_linux_ipc_actor_sys_bindings_cpp_4
{
private:
  static inline unsigned int hash (const char *str, size_t len);
public:
  static const struct word_type_gperf__src_linux_ipc_actor_sys_bindings_cpp_4 *in_word_set (const char *str, size_t len);
};

inline /*ARGSUSED*/
unsigned int
gperf__src_linux_ipc_actor_sys_bindings_cpp_4::hash (const char *str, size_t len)
{
  return len;
}

const struct word_type_gperf__src_linux_ipc_actor_sys_bindings_cpp_4 *
gperf__src_linux_ipc_actor_sys_bindings_cpp_4::in_word_set (const char *str, size_t len)
{
  enum
    {
      TOTAL_KEYWORDS = 2,
      MIN_WORD_LENGTH = 9,
      MAX_WORD_LENGTH = 14,
      MIN_HASH_VALUE = 9,
      MAX_HASH_VALUE = 14
    };

  static const struct word_type_gperf__src_linux_ipc_actor_sys_bindings_cpp_4 wordlist[] =
    {
      {"",{}}, {"",{}}, {"",{}}, {"",{}}, {"",{}}, {"",{}},
      {"",{}}, {"",{}}, {"",{}},
      {"parent_fd", 
                                [](
                                    lua_State* L, landlock_path_beneath_attr& at
                                ) {
                                    if (lua_type(L, -1) != LUA_TNUMBER) {
                                        return "invalid parent_fd";
                                    }

                                    at.parent_fd = lua_tointeger(L, -1);
                                    return (const char*)(NULL);
                                }},
      {"",{}}, {"",{}}, {"",{}}, {"",{}},
      {"allowed_access", 
                                [](
                                    lua_State* L, landlock_path_beneath_attr& at
                                ) {
                                    if (lua_type(L, -1) != LUA_TTABLE) {
                                        return "invalid allowed_access";
                                    }

                                    auto res = landlock_handled_access_fs(L);
                                    if (res) {
                                        at.allowed_access = res.value();
                                        return (const char*)(NULL);
                                    } else {
                                        return res.error();
                                    }
                                }}
    };

  if (len <= MAX_WORD_LENGTH && len >= MIN_WORD_LENGTH)
    {
      unsigned int key = hash (str, len);

      if (key <= MAX_HASH_VALUE)
        {
          const char *s = wordlist[key].name;

          if (*str == *s && !strcmp (str + 1, s + 1))
            return &wordlist[key];
        }
    }
  return 0;
}

} // namespace emilua::gperf::detail
namespace emilua::gperf::detail {
/* C++ code produced by gperf version 3.1 */
/* Command-line: /usr/bin/gperf --language=C++ --enum --readonly-tables --struct-type --initializer-suffix=',{}' --class-name=gperf__src_linux_ipc_actor_sys_bindings_cpp_1  */
/* Computed positions: -k'2-5,7,12,14' */

#if !((' ' == 32) && ('!' == 33) && ('"' == 34) && ('#' == 35) \
      && ('%' == 37) && ('&' == 38) && ('\'' == 39) && ('(' == 40) \
      && (')' == 41) && ('*' == 42) && ('+' == 43) && (',' == 44) \
      && ('-' == 45) && ('.' == 46) && ('/' == 47) && ('0' == 48) \
      && ('1' == 49) && ('2' == 50) && ('3' == 51) && ('4' == 52) \
      && ('5' == 53) && ('6' == 54) && ('7' == 55) && ('8' == 56) \
      && ('9' == 57) && (':' == 58) && (';' == 59) && ('<' == 60) \
      && ('=' == 61) && ('>' == 62) && ('?' == 63) && ('A' == 65) \
      && ('B' == 66) && ('C' == 67) && ('D' == 68) && ('E' == 69) \
      && ('F' == 70) && ('G' == 71) && ('H' == 72) && ('I' == 73) \
      && ('J' == 74) && ('K' == 75) && ('L' == 76) && ('M' == 77) \
      && ('N' == 78) && ('O' == 79) && ('P' == 80) && ('Q' == 81) \
      && ('R' == 82) && ('S' == 83) && ('T' == 84) && ('U' == 85) \
      && ('V' == 86) && ('W' == 87) && ('X' == 88) && ('Y' == 89) \
      && ('Z' == 90) && ('[' == 91) && ('\\' == 92) && (']' == 93) \
      && ('^' == 94) && ('_' == 95) && ('a' == 97) && ('b' == 98) \
      && ('c' == 99) && ('d' == 100) && ('e' == 101) && ('f' == 102) \
      && ('g' == 103) && ('h' == 104) && ('i' == 105) && ('j' == 106) \
      && ('k' == 107) && ('l' == 108) && ('m' == 109) && ('n' == 110) \
      && ('o' == 111) && ('p' == 112) && ('q' == 113) && ('r' == 114) \
      && ('s' == 115) && ('t' == 116) && ('u' == 117) && ('v' == 118) \
      && ('w' == 119) && ('x' == 120) && ('y' == 121) && ('z' == 122) \
      && ('{' == 123) && ('|' == 124) && ('}' == 125) && ('~' == 126))
/* The character set is not based on ISO-646.  */
#error "gperf generated tables don't work with this execution character set. Please report a bug to <bug-gperf@gnu.org>."
#endif

struct word_type_gperf__src_linux_ipc_actor_sys_bindings_cpp_1 { const char* name; int (*action)(lua_State*); };
/* maximum key range = 587, duplicates = 0 */

class gperf__src_linux_ipc_actor_sys_bindings_cpp_1
{
private:
  static inline unsigned int hash (const char *str, size_t len);
public:
  static const struct word_type_gperf__src_linux_ipc_actor_sys_bindings_cpp_1 *in_word_set (const char *str, size_t len);
};

inline unsigned int
gperf__src_linux_ipc_actor_sys_bindings_cpp_1::hash (const char *str, size_t len)
{
  static const unsigned short asso_values[] =
    {
      590, 590, 590, 590, 590, 590, 590, 590, 590, 590,
      590, 590, 590, 590, 590, 590, 590, 590, 590, 590,
      590, 590, 590, 590, 590, 590, 590, 590, 590, 590,
      590, 590, 590, 590, 590, 590, 590, 590, 590, 590,
      590, 590, 590, 590, 590, 590, 590, 590, 590, 590,
       15,  45, 590, 590, 590, 590, 590, 590, 590, 590,
      590, 590, 590, 590, 590, 140,  45, 175,   5,  10,
      115,  25, 220,   0,  15,  10,  50,  70,   0,   0,
       25,  84,  20,   0,  10,  80, 140, 214, 180, 245,
        0,  15, 590, 590, 590,   5,  20,   5, 105,  10,
        0,   0,  20,   5,  20,  90,  10, 200,  10, 170,
        0,   0,   0,   0,  55,   0,   0,   5,  25,   0,
      110,   0, 590, 590, 590, 590, 590, 590, 590, 590,
      590, 590, 590, 590, 590, 590, 590, 590, 590, 590,
      590, 590, 590, 590, 590, 590, 590, 590, 590, 590,
      590, 590, 590, 590, 590, 590, 590, 590, 590, 590,
      590, 590, 590, 590, 590, 590, 590, 590, 590, 590,
      590, 590, 590, 590, 590, 590, 590, 590, 590, 590,
      590, 590, 590, 590, 590, 590, 590, 590, 590, 590,
      590, 590, 590, 590, 590, 590, 590, 590, 590, 590,
      590, 590, 590, 590, 590, 590, 590, 590, 590, 590,
      590, 590, 590, 590, 590, 590, 590, 590, 590, 590,
      590, 590, 590, 590, 590, 590, 590, 590, 590, 590,
      590, 590, 590, 590, 590, 590, 590, 590, 590, 590,
      590, 590, 590, 590, 590, 590, 590, 590, 590, 590,
      590, 590, 590, 590, 590, 590, 590
    };
  unsigned int hval = len;

  switch (hval)
    {
      default:
        hval += asso_values[static_cast<unsigned char>(str[13])];
      /*FALLTHROUGH*/
      case 13:
      case 12:
        hval += asso_values[static_cast<unsigned char>(str[11])];
      /*FALLTHROUGH*/
      case 11:
      case 10:
      case 9:
      case 8:
      case 7:
        hval += asso_values[static_cast<unsigned char>(str[6]+1)];
      /*FALLTHROUGH*/
      case 6:
      case 5:
        hval += asso_values[static_cast<unsigned char>(str[4])];
      /*FALLTHROUGH*/
      case 4:
        hval += asso_values[static_cast<unsigned char>(str[3])];
      /*FALLTHROUGH*/
      case 3:
        hval += asso_values[static_cast<unsigned char>(str[2])];
      /*FALLTHROUGH*/
      case 2:
        hval += asso_values[static_cast<unsigned char>(str[1])];
        break;
    }
  return hval;
}

const struct word_type_gperf__src_linux_ipc_actor_sys_bindings_cpp_1 *
gperf__src_linux_ipc_actor_sys_bindings_cpp_1::in_word_set (const char *str, size_t len)
{
  enum
    {
      TOTAL_KEYWORDS = 239,
      MIN_WORD_LENGTH = 3,
      MAX_WORD_LENGTH = 34,
      MIN_HASH_VALUE = 3,
      MAX_HASH_VALUE = 589
    };

  static const struct word_type_gperf__src_linux_ipc_actor_sys_bindings_cpp_1 wordlist[] =
    {
      {"",{}}, {"",{}}, {"",{}},
      {"EIO",  EMILUA_DETAIL_INT_CONSTANT(EIO)},
      {"open", 
            [](lua_State* L) -> int {
                lua_pushcfunction(L, [](lua_State* L) -> int {
                    lua_settop(L, 3);
                    const char* str = luaL_checkstring(L, 1);
                    int flags = luaL_checkinteger(L, 2);
                    int res;
                    if (lua_isnil(L, 3)) {
                        res = open(str, flags);
                    } else {
                        mode_t mode = luaL_checkinteger(L, 3);
                        res = open(str, flags, mode);
                    }
                    int last_error = (res == -1) ? errno : 0;
                    CHECK_LAST_ERROR(L, last_error, "open");
                    lua_pushinteger(L, res);
                    lua_pushinteger(L, last_error);
                    return 2;
                });
                return 1;
            }},
      {"setns", 
            [](lua_State* L) -> int {
                lua_pushcfunction(L, [](lua_State* L) -> int {
                    int fd = luaL_checkinteger(L, 1);
                    int nstype = luaL_checkinteger(L, 2);
                    int res = setns(fd, nstype);
                    int last_error = (res == -1) ? errno : 0;
                    CHECK_LAST_ERROR(L, last_error, "setns");
                    lua_pushinteger(L, res);
                    lua_pushinteger(L, last_error);
                    return 2;
                });
                return 1;
            }},
      {"fsopen", 
            [](lua_State* L) -> int {
                lua_pushcfunction(L, [](lua_State* L) -> int {
                    const char* pathname;
                    unsigned int flags = luaL_checkinteger(L, 2);

                    switch (lua_type(L, 1)) {
                    case LUA_TSTRING:
                        pathname = lua_tostring(L, 1);
                        break;
                    case LUA_TNIL:
                        pathname = NULL;
                        break;
                    default:
                        errno = EINVAL;
                        perror("<3>ipc_actor/init/fsopen");
                        std::exit(1);
                    }

                    int res = fsopen(pathname, flags);
                    int last_error = (res == -1) ? errno : 0;
                    CHECK_LAST_ERROR(L, last_error, "fsopen");
                    lua_pushinteger(L, res);
                    lua_pushinteger(L, last_error);
                    return 2;
                });
                return 1;
            }},
      {"",{}}, {"",{}},
      {"read", 
            [](lua_State* L) -> int {
                lua_pushcfunction(L, [](lua_State* L) -> int {
                    int fd = luaL_checkinteger(L, 1);
                    int nbyte = luaL_checkinteger(L, 2);
                    void* ud;
                    lua_Alloc a = lua_getallocf(L, &ud);
                    char* buf = static_cast<char*>(a(ud, NULL, 0, nbyte));
                    int res = read(fd, buf, nbyte);
                    int last_error = (res == -1) ? errno : 0;
                    CHECK_LAST_ERROR(L, last_error, "read");
                    if (last_error == 0) {
                        lua_pushlstring(L, buf, res);
                    } else {
                        lua_pushnil(L);
                    }
                    lua_pushinteger(L, last_error);
                    return 2;
                });
                return 1;
            }},
      {"mount", 
            [](lua_State* L) -> int {
                lua_pushcfunction(L, [](lua_State* L) -> int {
                    const char* source =
                        lua_isnil(L, 1) ? nullptr : lua_tostring(L, 1);
                    const char* target = luaL_checkstring(L, 2);
                    const char* fstype =
                        lua_isnil(L, 3) ? nullptr : lua_tostring(L, 3);
                    unsigned long flags = luaL_checkinteger(L, 4);
                    const void* data =
                        lua_isnil(L, 5) ? nullptr : lua_tostring(L, 5);
                    int res = mount(source, target, fstype, flags, data);
                    int last_error = (res == -1) ? errno : 0;
                    CHECK_LAST_ERROR(L, last_error, "mount");
                    lua_pushinteger(L, res);
                    lua_pushinteger(L, last_error);
                    return 2;
                });
                return 1;
            }},
      {"EISDIR",  EMILUA_DETAIL_INT_CONSTANT(EISDIR)},
      {"setpgid", 
            [](lua_State* L) -> int {
                lua_pushcfunction(L, [](lua_State* L) -> int {
                    pid_t pid = luaL_checkinteger(L, 1);
                    pid_t pgid = luaL_checkinteger(L, 2);
                    int res = setpgid(pid, pgid);
                    int last_error = (res == -1) ? errno : 0;
                    CHECK_LAST_ERROR(L, last_error, "setpgid");
                    lua_pushinteger(L, res);
                    lua_pushinteger(L, last_error);
                    return 2;
                });
                return 1;
            }},
      {"",{}},
      {"open_tree", 
            [](lua_State* L) -> int {
                lua_pushcfunction(L, [](lua_State* L) -> int {
                    int dirfd = luaL_checkinteger(L, 1);
                    const char* pathname;
                    unsigned int flags = luaL_checkinteger(L, 3);

                    switch (lua_type(L, 2)) {
                    case LUA_TSTRING:
                        pathname = lua_tostring(L, 2);
                        break;
                    case LUA_TNIL:
                        pathname = NULL;
                        break;
                    default:
                        errno = EINVAL;
                        perror("<3>ipc_actor/init/open_tree");
                        std::exit(1);
                    }

                    int res = open_tree(dirfd, pathname, flags);
                    int last_error = (res == -1) ? errno : 0;
                    CHECK_LAST_ERROR(L, last_error, "open_tree");
                    lua_pushinteger(L, res);
                    lua_pushinteger(L, last_error);
                    return 2;
                });
                return 1;
            }},
      {"",{}},
      {"ENOENT",  EMILUA_DETAIL_INT_CONSTANT(ENOENT)},
      {"",{}},
      {"mount_setattr", 
            [](lua_State* L) -> int {
                lua_pushcfunction(L, [](lua_State* L) -> int {
                    int dirfd = luaL_checkinteger(L, 1);
                    const char* pathname;
                    unsigned int flags = luaL_checkinteger(L, 3);
                    luaL_checktype(L, 4, LUA_TTABLE);

                    switch (lua_type(L, 2)) {
                    case LUA_TSTRING:
                        pathname = lua_tostring(L, 2);
                        break;
                    case LUA_TNIL:
                        pathname = NULL;
                        break;
                    default:
                        errno = EINVAL;
                        perror("<3>ipc_actor/init/mount_setattr");
                        std::exit(1);
                    }

                    struct mount_attr attr;
                    std::memset(&attr, 0, sizeof(attr));

                    lua_pushliteral(L, "attr_set");
                    lua_rawget(L, 4);
                    switch (lua_type(L, -1)) {
                    case LUA_TNUMBER:
                        attr.attr_set = lua_tointeger(L, -1);
                        break;
                    case LUA_TNIL:
                        break;
                    default:
                        errno = EINVAL;
                        perror("<3>ipc_actor/init/mount_setattr/attr_set");
                        std::exit(1);
                    }
                    lua_pop(L, 1);

                    lua_pushliteral(L, "attr_clr");
                    lua_rawget(L, 4);
                    switch (lua_type(L, -1)) {
                    case LUA_TNUMBER:
                        attr.attr_clr = lua_tointeger(L, -1);
                        break;
                    case LUA_TNIL:
                        break;
                    default:
                        errno = EINVAL;
                        perror("<3>ipc_actor/init/mount_setattr/attr_clr");
                        std::exit(1);
                    }
                    lua_pop(L, 1);

                    lua_pushliteral(L, "propagation");
                    lua_rawget(L, 4);
                    switch (lua_type(L, -1)) {
                    case LUA_TNUMBER:
                        attr.propagation = lua_tointeger(L, -1);
                        break;
                    case LUA_TNIL:
                        break;
                    default:
                        errno = EINVAL;
                        perror("<3>ipc_actor/init/mount_setattr/propagation");
                        std::exit(1);
                    }
                    lua_pop(L, 1);

                    lua_pushliteral(L, "userns_fd");
                    lua_rawget(L, 4);
                    switch (lua_type(L, -1)) {
                    case LUA_TNUMBER:
                        attr.userns_fd = lua_tointeger(L, -1);
                        break;
                    case LUA_TNIL:
                        break;
                    default:
                        errno = EINVAL;
                        perror("<3>ipc_actor/init/mount_setattr/userns_fd");
                        std::exit(1);
                    }
                    lua_pop(L, 1);

                    int res = mount_setattr(
                        dirfd, pathname, flags, &attr, sizeof(attr));
                    int last_error = (res == -1) ? errno : 0;
                    CHECK_LAST_ERROR(L, last_error, "mount_setattr");
                    lua_pushinteger(L, res);
                    lua_pushinteger(L, last_error);
                    return 2;
                });
                return 1;
            }},
      {"",{}}, {"",{}},
      {"ENODEV",  EMILUA_DETAIL_INT_CONSTANT(ENODEV)},
      {"ENOTDIR",  EMILUA_DETAIL_INT_CONSTANT(ENOTDIR)},
      {"ENOTSOCK",  EMILUA_DETAIL_INT_CONSTANT(ENOTSOCK)},
      {"",{}},
      {"chown", 
            [](lua_State* L) -> int {
                lua_pushcfunction(L, [](lua_State* L) -> int {
                    const char* path = luaL_checkstring(L, 1);
                    uid_t owner = luaL_checkinteger(L, 2);
                    gid_t group = luaL_checkinteger(L, 3);
                    int res = chown(path, owner, group);
                    int last_error = (res == -1) ? errno : 0;
                    CHECK_LAST_ERROR(L, last_error, "chown");
                    lua_pushinteger(L, res);
                    lua_pushinteger(L, last_error);
                    return 2;
                });
                return 1;
            }},
      {"ENOTTY",  EMILUA_DETAIL_INT_CONSTANT(ENOTTY)},
      {"",{}},
      {"fsconfig", 
            [](lua_State* L) -> int {
                lua_pushcfunction(L, [](lua_State* L) -> int {
                    int fd = luaL_checkinteger(L, 1);
                    int cmd = luaL_checkinteger(L, 2);
                    const char* key;
                    const char* value;
                    int aux = luaL_checkinteger(L, 5);

                    switch (lua_type(L, 3)) {
                    case LUA_TSTRING:
                        key = lua_tostring(L, 3);
                        break;
                    case LUA_TNIL:
                        key = NULL;
                        break;
                    default:
                        errno = EINVAL;
                        perror("<3>ipc_actor/init/fsconfig");
                        std::exit(1);
                    }

                    switch (lua_type(L, 4)) {
                    case LUA_TSTRING:
                        value = lua_tostring(L, 4);
                        break;
                    case LUA_TNIL:
                        value = NULL;
                        break;
                    default:
                        errno = EINVAL;
                        perror("<3>ipc_actor/init/fsconfig");
                        std::exit(1);
                    }

                    int res = fsconfig(fd, cmd, key, value, aux);
                    int last_error = (res == -1) ? errno : 0;
                    CHECK_LAST_ERROR(L, last_error, "fsconfig");
                    lua_pushinteger(L, res);
                    lua_pushinteger(L, last_error);
                    return 2;
                });
                return 1;
            }},
      {"cap_drop_bound", 
            [](lua_State* L) -> int {
                lua_pushcfunction(L, [](lua_State* L) -> int {
                    const char* name = luaL_checkstring(L, 1);
                    cap_value_t cap;
                    if (cap_from_name(name, &cap) == -1) {
                        perror("<3>ipc_actor/init/cap_drop_bound");
                        std::exit(1);
                    }
                    int res = cap_drop_bound(cap);
                    int last_error = (res == -1) ? errno : 0;
                    CHECK_LAST_ERROR(L, last_error, "cap_drop_bound");
                    lua_pushinteger(L, res);
                    lua_pushinteger(L, last_error);
                    return 2;
                });
                return 1;
            }},
      {"MS_DIRSYNC",  EMILUA_DETAIL_INT_CONSTANT(MS_DIRSYNC)},
      {"ENOSPC",  EMILUA_DETAIL_INT_CONSTANT(ENOSPC)},
      {"S_IROTH",  EMILUA_DETAIL_INT_CONSTANT(S_IROTH)},
      {"",{}}, {"",{}},
      {"EINTR",  EMILUA_DETAIL_INT_CONSTANT(EINTR)},
      {"sethostname", 
            [](lua_State* L) -> int {
                lua_pushcfunction(L, [](lua_State* L) -> int {
                    std::size_t len;
                    const char* str = lua_tolstring(L, 1, &len);
                    int res = sethostname(str, len);
                    int last_error = (res == -1) ? errno : 0;
                    CHECK_LAST_ERROR(L, last_error, "sethostname");
                    lua_pushinteger(L, res);
                    lua_pushinteger(L, last_error);
                    return 2;
                });
                return 1;
            }},
      {"cap_set_proc", 
            [](lua_State* L) -> int {
                lua_pushcfunction(L, [](lua_State* L) -> int {
                    const char* caps = luaL_checkstring(L, 1);
                    cap_t caps2 = cap_from_text(caps);
                    if (caps2 == NULL) {
                        perror("<3>ipc_actor/init/cap_set_proc");
                        std::exit(1);
                    }
                    BOOST_SCOPE_EXIT_ALL(&) { cap_free(caps2); };
                    int res = cap_set_proc(caps2);
                    int last_error = (res == -1) ? errno : 0;
                    CHECK_LAST_ERROR(L, last_error, "cap_set_proc");
                    lua_pushinteger(L, res);
                    lua_pushinteger(L, last_error);
                    return 2;
                });
                return 1;
            }},
      {"landlock_create_ruleset", 
            [](lua_State* L) -> int {
                lua_pushcfunction(L, [](lua_State* L) -> int {
                    lua_settop(L, 2);

                    bool has_ruleset;
                    bool has_flags;

                    switch (lua_type(L, 1)) {
                    case LUA_TTABLE:
                        has_ruleset = true;
                        break;
                    case LUA_TNIL:
                        has_ruleset = false;
                        break;
                    default:
                        return luaL_error(L, "table expected for argument 1");
                    }

                    switch (lua_type(L, 2)) {
                    case LUA_TTABLE:
                        has_flags = true;
                        break;
                    case LUA_TNIL:
                        has_flags = false;
                        break;
                    default:
                        return luaL_error(L, "table expected for argument 2");
                    }

                    struct landlock_ruleset_attr ruleset_attr;
                    std::memset(&ruleset_attr, 0, sizeof(ruleset_attr));

                    if (has_ruleset) {
                        lua_pushnil(L);
                        while (lua_next(L, 1) != 0) {
                            if (lua_type(L, -2) != LUA_TSTRING) {
                                return luaL_error(L, "invalid ruleset attr");
                            }

                            auto hkey = tostringview(L, -2);
                            auto errstr =  ::emilua::gperf::detail::value_or(::emilua::gperf::detail::gperf__src_linux_ipc_actor_sys_bindings_cpp_2::in_word_set((hkey).data(), (hkey).size()), 
                                    [](lua_State*, landlock_ruleset_attr&) {
                                        return "invalid ruleset attr";
                                    })(L, ruleset_attr);
                            if (errstr) {
                                lua_pushstring(L, errstr);
                                return lua_error(L);
                            }
                            lua_pop(L, 1);
                        }
                    }

                    std::uint32_t flags = 0;
                    if (has_flags) {
                        for (int i = 0 ;; ++i) {
                            lua_rawgeti(L, 2, i + 1);
                            switch (lua_type(L, -1)) {
                            case LUA_TNIL: {
                                lua_pop(L, 1);
                                goto end_for;
                            }
                            case LUA_TSTRING:
                                break;
                            default:
                                return luaL_error(
                                    L, "invalid LANDLOCK_CREATE_RULESET flag");
                            }

                            auto fkey = tostringview(L, -1);
                            auto flag =  ::emilua::gperf::detail::value_or(::emilua::gperf::detail::gperf__src_linux_ipc_actor_sys_bindings_cpp_3::in_word_set((fkey).data(), (fkey).size()), 0);
                            if (flag == 0) {
                                return luaL_error(
                                    L, "invalid LANDLOCK_CREATE_RULESET flag");
                            }
                            flags |= flag;
                            lua_pop(L, 1);
                        }
                    end_for:;
                    }

                    int res = syscall(SYS_landlock_create_ruleset,
                                      has_ruleset ? &ruleset_attr : NULL,
                                      has_ruleset ? sizeof(ruleset_attr) : 0,
                                      flags);

                    int last_error = (res == -1) ? errno : 0;
                    CHECK_LAST_ERROR(L, last_error, "landlock_create_ruleset");
                    lua_pushinteger(L, res);
                    lua_pushinteger(L, last_error);
                    return 2;
                });
                return 1;
            }},
      {"MS_RDONLY",  EMILUA_DETAIL_INT_CONSTANT(MS_RDONLY)},
      {"move_mount", 
            [](lua_State* L) -> int {
                lua_pushcfunction(L, [](lua_State* L) -> int {
                    int fdirfd = luaL_checkinteger(L, 1);
                    int tdirfd = luaL_checkinteger(L, 3);
                    const char* fpath;
                    const char* tpath;
                    unsigned int flags = luaL_checkinteger(L, 5);

                    switch (lua_type(L, 2)) {
                    case LUA_TSTRING:
                        fpath = lua_tostring(L, 2);
                        break;
                    case LUA_TNIL:
                        fpath = NULL;
                        break;
                    default:
                        errno = EINVAL;
                        perror("<3>ipc_actor/init/move_mount");
                        std::exit(1);
                    }

                    switch (lua_type(L, 4)) {
                    case LUA_TSTRING:
                        tpath = lua_tostring(L, 4);
                        break;
                    case LUA_TNIL:
                        tpath = NULL;
                        break;
                    default:
                        errno = EINVAL;
                        perror("<3>ipc_actor/init/move_mount");
                        std::exit(1);
                    }

                    int res = move_mount(fdirfd, fpath, tdirfd, tpath, flags);
                    int last_error = (res == -1) ? errno : 0;
                    CHECK_LAST_ERROR(L, last_error, "move_mount");
                    lua_pushinteger(L, res);
                    lua_pushinteger(L, last_error);
                    return 2;
                });
                return 1;
            }},
      {"MS_REC",  EMILUA_DETAIL_INT_CONSTANT(MS_REC)},
      {"",{}},
      {"O_DIRECT",  EMILUA_DETAIL_INT_CONSTANT(O_DIRECT)},
      {"",{}},
      {"EOPNOTSUPP",  EMILUA_DETAIL_INT_CONSTANT(EOPNOTSUPP)},
      {"O_DIRECTORY",  EMILUA_DETAIL_INT_CONSTANT(O_DIRECTORY)},
      {"S_ISGID",  EMILUA_DETAIL_INT_CONSTANT(S_ISGID)},
      {"seccomp_set_mode_filter", 
            [](lua_State* L) -> int {
                lua_pushcfunction(L, [](lua_State* L) -> int {
                    std::size_t len;
                    const char* data = lua_tolstring(L, 1, &len);

                    if (len == 0 || len % sizeof(struct sock_filter) != 0) {
                        CHECK_LAST_ERROR(L, EINVAL, "seccomp_set_mode_filter");
                        lua_pushinteger(L, -1);
                        lua_pushinteger(L, EINVAL);
                        return 2;
                    }

                    struct sock_fprog prog;
                    prog.len = len / sizeof(struct sock_filter);

                    if (
                        reinterpret_cast<std::uintptr_t>(data) %
                        alignof(sock_filter) == 0
                    ) {
                        // data already aligned
                        prog.filter = reinterpret_cast<sock_filter*>(
                            const_cast<char*>(data));
                    } else {
                        void* ud;
                        lua_Alloc a = lua_getallocf(L, &ud);
                        prog.filter =
                            static_cast<sock_filter*>(a(ud, NULL, 0, len));
                        std::memcpy(prog.filter, data, len);
                    }

                    int res = prctl(PR_SET_SECCOMP, SECCOMP_MODE_FILTER, &prog);
                    int last_error = (res == -1) ? errno : 0;
                    CHECK_LAST_ERROR(L, last_error, "seccomp_set_mode_filter");
                    lua_pushinteger(L, res);
                    lua_pushinteger(L, last_error);
                    return 2;
                });
                return 1;
            }},
      {"",{}}, {"",{}}, {"",{}},
      {"unshare", 
            [](lua_State* L) -> int {
                lua_pushcfunction(L, [](lua_State* L) -> int {
                    int nstype = luaL_checkinteger(L, 1);
                    int res = unshare(nstype);
                    int last_error = (res == -1) ? errno : 0;
                    CHECK_LAST_ERROR(L, last_error, "unshare");
                    lua_pushinteger(L, res);
                    lua_pushinteger(L, last_error);
                    return 2;
                });
                return 1;
            }},
      {"",{}},
      {"MS_STRICTATIME",  EMILUA_DETAIL_INT_CONSTANT(MS_STRICTATIME)},
      {"",{}},
      {"ESPIPE",  EMILUA_DETAIL_INT_CONSTANT(ESPIPE)},
      {"",{}}, {"",{}},
      {"ENETRESET",  EMILUA_DETAIL_INT_CONSTANT(ENETRESET)},
      {"",{}},
      {"EPROTO",  EMILUA_DETAIL_INT_CONSTANT(EPROTO)},
      {"",{}}, {"",{}}, {"",{}},
      {"EPIPE",  EMILUA_DETAIL_INT_CONSTANT(EPIPE)},
      {"EILSEQ",  EMILUA_DETAIL_INT_CONSTANT(EILSEQ)},
      {"MS_BIND",  EMILUA_DETAIL_INT_CONSTANT(MS_BIND)},
      {"",{}}, {"",{}},
      {"MS_REMOUNT",  EMILUA_DETAIL_INT_CONSTANT(MS_REMOUNT)},
      {"CLONE_NEWNS",  EMILUA_DETAIL_INT_CONSTANT(CLONE_NEWNS)},
      {"CLONE_NEWUTS",  EMILUA_DETAIL_INT_CONSTANT(CLONE_NEWUTS)},
      {"",{}}, {"",{}},
      {"AT_NO_AUTOMOUNT",  EMILUA_DETAIL_INT_CONSTANT(AT_NO_AUTOMOUNT)},
      {"ENOMSG",  EMILUA_DETAIL_INT_CONSTANT(ENOMSG)},
      {"CLONE_NEWPID",  EMILUA_DETAIL_INT_CONSTANT(CLONE_NEWPID)},
      {"",{}},
      {"EDOM",  EMILUA_DETAIL_INT_CONSTANT(EDOM)},
      {"ELOOP",  EMILUA_DETAIL_INT_CONSTANT(ELOOP)},
      {"chroot", 
            [](lua_State* L) -> int {
                lua_pushcfunction(L, [](lua_State* L) -> int {
                    const char* path = luaL_checkstring(L, 1);
                    int res = chroot(path);
                    int last_error = (res == -1) ? errno : 0;
                    CHECK_LAST_ERROR(L, last_error, "chroot");
                    lua_pushinteger(L, res);
                    lua_pushinteger(L, last_error);
                    return 2;
                });
                return 1;
            }},
      {"CLONE_NEWNET",  EMILUA_DETAIL_INT_CONSTANT(CLONE_NEWNET)},
      {"CLONE_NEWUSER",  EMILUA_DETAIL_INT_CONSTANT(CLONE_NEWUSER)},
      {"setresgid", 
            [](lua_State* L) -> int {
                lua_pushcfunction(L, [](lua_State* L) -> int {
                    gid_t rgid = luaL_checkinteger(L, 1);
                    gid_t egid = luaL_checkinteger(L, 2);
                    gid_t sgid = luaL_checkinteger(L, 3);
                    int res = setresgid(rgid, egid, sgid);
                    int last_error = (res == -1) ? errno : 0;
                    CHECK_LAST_ERROR(L, last_error, "setresgid");
                    lua_pushinteger(L, res);
                    lua_pushinteger(L, last_error);
                    return 2;
                });
                return 1;
            }},
      {"O_NONBLOCK",  EMILUA_DETAIL_INT_CONSTANT(O_NONBLOCK)},
      {"ENOMEM",  EMILUA_DETAIL_INT_CONSTANT(ENOMEM)},
      {"landlock_add_rule", 
            [](lua_State* L) -> int {
                lua_pushcfunction(L, [](lua_State* L) -> int {
                    lua_settop(L, 4);

                    switch (lua_type(L, 1)) {
                    case LUA_TNUMBER:
                        break;
                    default:
                        return luaL_error(L, "integer expected for argument 1");
                    }

                    if (tostringview(L, 2) != "path_beneath") {
                        return luaL_error(L, "invalid argument 2");
                    }

                    if (lua_type(L, 3) != LUA_TTABLE) {
                        return luaL_error(L, "table expected for argument 3");
                    }

                    switch (lua_type(L, 4)) {
                    case LUA_TNIL:
                        break;
                    default:
                        return luaL_error(L, "nil expected for argument 4");
                    }

                    struct landlock_path_beneath_attr path_beneath_attr;
                    std::memset(
                        &path_beneath_attr, 0, sizeof(path_beneath_attr));
                    path_beneath_attr.parent_fd = -1;

                    lua_pushnil(L);
                    while (lua_next(L, 3) != 0) {
                        if (lua_type(L, -2) != LUA_TSTRING) {
                            return luaL_error(L, "invalid path_beneath attr");
                        }

                        auto hkey = tostringview(L, -2);
                        auto errstr =  ::emilua::gperf::detail::value_or(::emilua::gperf::detail::gperf__src_linux_ipc_actor_sys_bindings_cpp_4::in_word_set((hkey).data(), (hkey).size()), 
                                [](lua_State*, landlock_path_beneath_attr&) {
                                    return "invalid path_beneath attr";
                                })(L, path_beneath_attr);
                        if (errstr) {
                            lua_pushstring(L, errstr);
                            return lua_error(L);
                        }
                        lua_pop(L, 1);
                    }

                    int res = syscall(
                        SYS_landlock_add_rule,
                        lua_tointeger(L, 1),
                        /*LANDLOCK_RULE_PATH_BENEATH=*/1,
                        &path_beneath_attr,
                        /*flags=*/0);

                    int last_error = (res == -1) ? errno : 0;
                    CHECK_LAST_ERROR(L, last_error, "landlock_add_rule");
                    lua_pushinteger(L, res);
                    lua_pushinteger(L, last_error);
                    return 2;
                });
                return 1;
            }},
      {"",{}},
      {"setresuid", 
            [](lua_State* L) -> int {
                lua_pushcfunction(L, [](lua_State* L) -> int {
                    uid_t ruid = luaL_checkinteger(L, 1);
                    uid_t euid = luaL_checkinteger(L, 2);
                    uid_t suid = luaL_checkinteger(L, 3);
                    int res = setresuid(ruid, euid, suid);
                    int last_error = (res == -1) ? errno : 0;
                    CHECK_LAST_ERROR(L, last_error, "setresuid");
                    lua_pushinteger(L, res);
                    lua_pushinteger(L, last_error);
                    return 2;
                });
                return 1;
            }},
      {"E2BIG",  EMILUA_DETAIL_INT_CONSTANT(E2BIG)},
      {"MS_RELATIME",  EMILUA_DETAIL_INT_CONSTANT(MS_RELATIME)},
      {"landlock_restrict_self", 
            [](lua_State* L) -> int {
                lua_pushcfunction(L, [](lua_State* L) -> int {
                    lua_settop(L, 2);

                    switch (lua_type(L, 1)) {
                    case LUA_TNUMBER:
                        break;
                    default:
                        return luaL_error(L, "integer expected for argument 1");
                    }

                    switch (lua_type(L, 2)) {
                    case LUA_TNIL:
                        break;
                    default:
                        return luaL_error(L, "nil expected for argument 2");
                    }

                    int res = syscall(
                        SYS_landlock_restrict_self, lua_tointeger(L, 1), 0);

                    int last_error = (res == -1) ? errno : 0;
                    CHECK_LAST_ERROR(L, last_error, "landlock_restrict_self");
                    lua_pushinteger(L, res);
                    lua_pushinteger(L, last_error);
                    return 2;
                });
                return 1;
            }},
      {"",{}},
      {"setgroups", 
            [](lua_State* L) -> int {
                lua_pushcfunction(L, [](lua_State* L) -> int {
                    luaL_checktype(L, 1, LUA_TTABLE);
                    std::vector<gid_t> groups;
                    for (int i = 0 ;; ++i) {
                        lua_rawgeti(L, 1, i + 1);
                        switch (lua_type(L, -1)) {
                        case LUA_TNIL: {
                            int res = setgroups(groups.size(), groups.data());
                            int last_error = (res == -1) ? errno : 0;
                            CHECK_LAST_ERROR(L, last_error, "setgroups");
                            lua_pushinteger(L, res);
                            lua_pushinteger(L, last_error);
                            return 2;
                        }
                        case LUA_TNUMBER:
                            groups.emplace_back(lua_tointeger(L, -1));
                            lua_pop(L, 1);
                            break;
                        default:
                            errno = EINVAL;
                            perror("<3>ipc_actor/init/setgroups");
                            std::exit(1);
                        }
                    }
                });
                return 1;
            }},
      {"MS_NOATIME",  EMILUA_DETAIL_INT_CONSTANT(MS_NOATIME)},
      {"setsid", 
            [](lua_State* L) -> int {
                lua_pushcfunction(L, [](lua_State* L) -> int {
                    pid_t res = setsid();
                    int last_error = (res == -1) ? errno : 0;
                    CHECK_LAST_ERROR(L, last_error, "setsid");
                    lua_pushinteger(L, res);
                    lua_pushinteger(L, last_error);
                    return 2;
                });
                return 1;
            }},
      {"",{}}, {"",{}}, {"",{}},
      {"EIDRM",  EMILUA_DETAIL_INT_CONSTANT(EIDRM)},
      {"ENOTSUP",  EMILUA_DETAIL_INT_CONSTANT(ENOTSUP)},
      {"S_ISUID",  EMILUA_DETAIL_INT_CONSTANT(S_ISUID)},
      {"MS_NODIRATIME",  EMILUA_DETAIL_INT_CONSTANT(MS_NODIRATIME)},
      {"",{}},
      {"OPEN_TREE_CLONE",  EMILUA_DETAIL_INT_CONSTANT(OPEN_TREE_CLONE)},
      {"fspick", 
            [](lua_State* L) -> int {
                lua_pushcfunction(L, [](lua_State* L) -> int {
                    int dirfd = luaL_checkinteger(L, 1);
                    const char* pathname;
                    unsigned int flags = luaL_checkinteger(L, 3);

                    switch (lua_type(L, 2)) {
                    case LUA_TSTRING:
                        pathname = lua_tostring(L, 2);
                        break;
                    case LUA_TNIL:
                        pathname = NULL;
                        break;
                    default:
                        errno = EINVAL;
                        perror("<3>ipc_actor/init/fspick");
                        std::exit(1);
                    }

                    int res = fspick(dirfd, pathname, flags);
                    int last_error = (res == -1) ? errno : 0;
                    CHECK_LAST_ERROR(L, last_error, "fspick");
                    lua_pushinteger(L, res);
                    lua_pushinteger(L, last_error);
                    return 2;
                });
                return 1;
            }},
      {"ENOLINK",  EMILUA_DETAIL_INT_CONSTANT(ENOLINK)},
      {"O_RDONLY",  EMILUA_DETAIL_INT_CONSTANT(O_RDONLY)},
      {"",{}},
      {"MNT_DETACH",  EMILUA_DETAIL_INT_CONSTANT(MNT_DETACH)},
      {"ENETUNREACH",  EMILUA_DETAIL_INT_CONSTANT(ENETUNREACH)},
      {"S_IRUSR",  EMILUA_DETAIL_INT_CONSTANT(S_IRUSR)},
      {"ENOTEMPTY",  EMILUA_DETAIL_INT_CONSTANT(ENOTEMPTY)},
      {"",{}},
      {"EPROTONOSUPPORT",  EMILUA_DETAIL_INT_CONSTANT(EPROTONOSUPPORT)},
      {"",{}},
      {"OPEN_TREE_CLOEXEC",  EMILUA_DETAIL_INT_CONSTANT(OPEN_TREE_CLOEXEC)},
      {"EMSGSIZE",  EMILUA_DETAIL_INT_CONSTANT(EMSGSIZE)},
      {"MNT_EXPIRE",  EMILUA_DETAIL_INT_CONSTANT(MNT_EXPIRE)},
      {"cap_set_ambient", 
            [](lua_State* L) -> int {
                lua_pushcfunction(L, [](lua_State* L) -> int {
                    const char* name = luaL_checkstring(L, 1);
                    cap_value_t cap;
                    if (cap_from_name(name, &cap) == -1) {
                        perror("<3>ipc_actor/init/cap_set_ambient");
                        std::exit(1);
                    }

                    luaL_checktype(L, 2, LUA_TBOOLEAN);
                    cap_flag_value_t value =
                        lua_toboolean(L, 2) ? CAP_SET : CAP_CLEAR;

                    int res = cap_set_ambient(cap, value);
                    int last_error = (res == -1) ? errno : 0;
                    CHECK_LAST_ERROR(L, last_error, "cap_set_ambient");
                    lua_pushinteger(L, res);
                    lua_pushinteger(L, last_error);
                    return 2;
                });
                return 1;
            }},
      {"",{}}, {"",{}}, {"",{}},
      {"ETIMEDOUT",  EMILUA_DETAIL_INT_CONSTANT(ETIMEDOUT)},
      {"pivot_root", 
            [](lua_State* L) -> int {
                lua_pushcfunction(L, [](lua_State* L) -> int {
                    const char* new_root = luaL_checkstring(L, 1);
                    const char* put_old = luaL_checkstring(L, 2);
                    int res = syscall(SYS_pivot_root, new_root, put_old);
                    int last_error = (res == -1) ? errno : 0;
                    CHECK_LAST_ERROR(L, last_error, "pivot_root");
                    lua_pushinteger(L, res);
                    lua_pushinteger(L, last_error);
                    return 2;
                });
                return 1;
            }},
      {"EMLINK",  EMILUA_DETAIL_INT_CONSTANT(EMLINK)},
      {"O_TRUNC",  EMILUA_DETAIL_INT_CONSTANT(O_TRUNC)},
      {"MS_NODEV",  EMILUA_DETAIL_INT_CONSTANT(MS_NODEV)},
      {"MS_SILENT",  EMILUA_DETAIL_INT_CONSTANT(MS_SILENT)},
      {"EPERM",  EMILUA_DETAIL_INT_CONSTANT(EPERM)},
      {"EDESTADDRREQ",  EMILUA_DETAIL_INT_CONSTANT(EDESTADDRREQ)},
      {"",{}}, {"",{}},
      {"O_TMPFILE",  EMILUA_DETAIL_INT_CONSTANT(O_TMPFILE)},
      {"cap_set_secbits", 
            [](lua_State* L) -> int {
                lua_pushcfunction(L, [](lua_State* L) -> int {
                    int res = cap_set_secbits(luaL_checkinteger(L, 1));
                    int last_error = (res == -1) ? errno : 0;
                    CHECK_LAST_ERROR(L, last_error, "cap_set_secbits");
                    lua_pushinteger(L, res);
                    lua_pushinteger(L, last_error);
                    return 2;
                });
                return 1;
            }},
      {"ENOPROTOOPT",  EMILUA_DETAIL_INT_CONSTANT(ENOPROTOOPT)},
      {"",{}}, {"",{}},
      {"MNT_FORCE",  EMILUA_DETAIL_INT_CONSTANT(MNT_FORCE)},
      {"EROFS",  EMILUA_DETAIL_INT_CONSTANT(EROFS)},
      {"S_IRGRP",  EMILUA_DETAIL_INT_CONSTANT(S_IRGRP)},
      {"ENOBUFS",  EMILUA_DETAIL_INT_CONSTANT(ENOBUFS)},
      {"CLONE_NEWTIME",  EMILUA_DETAIL_INT_CONSTANT(CLONE_NEWTIME)},
      {"",{}},
      {"EPROTOTYPE",  EMILUA_DETAIL_INT_CONSTANT(EPROTOTYPE)},
      {"",{}},
      {"fexecve", 
            [](lua_State* L) -> int {
                lua_pushcfunction(L, [](lua_State* L) -> int {
                    lua_settop(L, 3);

                    int fd = luaL_checkinteger(L, 1);

                    std::vector<std::string> argvbuf;
                    switch (lua_type(L, 2)) {
                    case LUA_TNIL:
                        break;
                    case LUA_TTABLE:
                        for (int i = 1 ;; ++i) {
                            lua_rawgeti(L, 2, i);
                            switch (lua_type(L, -1)) {
                            case LUA_TNIL:
                                goto end_for;
                            case LUA_TSTRING:
                                argvbuf.emplace_back(tostringview(L));
                                lua_pop(L, 1);
                                break;
                            default:
                                errno = EINVAL;
                                perror("<3>ipc_actor/init/fexecve");
                                std::exit(1);
                            }
                        }
                    end_for:
                        break;
                    default:
                        errno = EINVAL;
                        perror("<3>ipc_actor/init/fexecve");
                        std::exit(1);
                    }

                    std::vector<std::string> envpbuf;
                    switch (lua_type(L, 3)) {
                    case LUA_TNIL:
                        break;
                    case LUA_TTABLE:
                        lua_pushnil(L);
                        while (lua_next(L, 3) != 0) {
                            if (
                                lua_type(L, -2) != LUA_TSTRING ||
                                lua_type(L, -1) != LUA_TSTRING
                            ) {
                                errno = EINVAL;
                                perror("<3>ipc_actor/init/fexecve");
                                std::exit(1);
                            }

                            envpbuf.emplace_back(tostringview(L, -2));
                            envpbuf.back() += '=';
                            envpbuf.back() += tostringview(L, -1);
                            lua_pop(L, 1);
                        }
                        break;
                    default:
                        errno = EINVAL;
                        perror("<3>ipc_actor/init/fexecve");
                        std::exit(1);
                    }

                    std::vector<char*> argv;
                    argv.reserve(argvbuf.size() + 1);
                    for (auto& a : argvbuf) {
                        argv.emplace_back(a.data());
                    }
                    argv.emplace_back(nullptr);

                    std::vector<char*> envp;
                    envp.reserve(envpbuf.size() + 1);
                    for (auto& e : envpbuf) {
                        envp.emplace_back(e.data());
                    }
                    envp.emplace_back(nullptr);

                    int res = fexecve(fd, argv.data(), envp.data());
                    int last_error = errno;
                    CHECK_LAST_ERROR(L, last_error, "fexecve");
                    lua_pushinteger(L, res);
                    lua_pushinteger(L, last_error);
                    return 2;
                });
                return 1;
            }},
      {"",{}}, {"",{}},
      {"write", 
            [](lua_State* L) -> int {
                lua_pushcfunction(L, [](lua_State* L) -> int {
                    int fd = luaL_checkinteger(L, 1);
                    std::size_t len;
                    const char* str = lua_tolstring(L, 2, &len);
                    int res = write(fd, str, len);
                    int last_error = (res == -1) ? errno : 0;
                    CHECK_LAST_ERROR(L, last_error, "write");
                    lua_pushinteger(L, res);
                    lua_pushinteger(L, last_error);
                    return 2;
                });
                return 1;
            }},
      {"execve", 
            [](lua_State* L) -> int {
                lua_pushcfunction(L, [](lua_State* L) -> int {
                    lua_settop(L, 3);

                    const char* pathname = luaL_checkstring(L, 1);

                    std::vector<std::string> argvbuf;
                    switch (lua_type(L, 2)) {
                    case LUA_TNIL:
                        break;
                    case LUA_TTABLE:
                        for (int i = 1 ;; ++i) {
                            lua_rawgeti(L, 2, i);
                            switch (lua_type(L, -1)) {
                            case LUA_TNIL:
                                goto end_for;
                            case LUA_TSTRING:
                                argvbuf.emplace_back(tostringview(L));
                                lua_pop(L, 1);
                                break;
                            default:
                                errno = EINVAL;
                                perror("<3>ipc_actor/init/execve");
                                std::exit(1);
                            }
                        }
                    end_for:
                        break;
                    default:
                        errno = EINVAL;
                        perror("<3>ipc_actor/init/execve");
                        std::exit(1);
                    }

                    std::vector<std::string> envpbuf;
                    switch (lua_type(L, 3)) {
                    case LUA_TNIL:
                        break;
                    case LUA_TTABLE:
                        lua_pushnil(L);
                        while (lua_next(L, 3) != 0) {
                            if (
                                lua_type(L, -2) != LUA_TSTRING ||
                                lua_type(L, -1) != LUA_TSTRING
                            ) {
                                errno = EINVAL;
                                perror("<3>ipc_actor/init/execve");
                                std::exit(1);
                            }

                            envpbuf.emplace_back(tostringview(L, -2));
                            envpbuf.back() += '=';
                            envpbuf.back() += tostringview(L, -1);
                            lua_pop(L, 1);
                        }
                        break;
                    default:
                        errno = EINVAL;
                        perror("<3>ipc_actor/init/execve");
                        std::exit(1);
                    }

                    std::vector<char*> argv;
                    argv.reserve(argvbuf.size() + 1);
                    for (auto& a : argvbuf) {
                        argv.emplace_back(a.data());
                    }
                    argv.emplace_back(nullptr);

                    std::vector<char*> envp;
                    envp.reserve(envpbuf.size() + 1);
                    for (auto& e : envpbuf) {
                        envp.emplace_back(e.data());
                    }
                    envp.emplace_back(nullptr);

                    int res = execve(pathname, argv.data(), envp.data());
                    int last_error = errno;
                    CHECK_LAST_ERROR(L, last_error, "execve");
                    lua_pushinteger(L, res);
                    lua_pushinteger(L, last_error);
                    return 2;
                });
                return 1;
            }},
      {"MOUNT_ATTR_NOSUID",  EMILUA_DETAIL_INT_CONSTANT(MOUNT_ATTR_NOSUID)},
      {"",{}},
      {"MS_NOSUID",  EMILUA_DETAIL_INT_CONSTANT(MS_NOSUID)},
      {"",{}},
      {"MOUNT_ATTR_NODEV",  EMILUA_DETAIL_INT_CONSTANT(MOUNT_ATTR_NODEV)},
      {"MOUNT_ATTR_NOSYMFOLLOW", 
            EMILUA_DETAIL_INT_CONSTANT(MOUNT_ATTR_NOSYMFOLLOW)},
      {"",{}}, {"",{}}, {"",{}},
      {"MOUNT_ATTR_NODIRATIME", 
            EMILUA_DETAIL_INT_CONSTANT(MOUNT_ATTR_NODIRATIME)},
      {"MOUNT_ATTR_NOEXEC",  EMILUA_DETAIL_INT_CONSTANT(MOUNT_ATTR_NOEXEC)},
      {"MS_UNBINDABLE",  EMILUA_DETAIL_INT_CONSTANT(MS_UNBINDABLE)},
      {"",{}}, {"",{}}, {"",{}},
      {"MOUNT_ATTR__ATIME",  EMILUA_DETAIL_INT_CONSTANT(MOUNT_ATTR__ATIME)},
      {"",{}},
      {"O_NOATIME",  EMILUA_DETAIL_INT_CONSTANT(O_NOATIME)},
      {"chdir", 
            [](lua_State* L) -> int {
                lua_pushcfunction(L, [](lua_State* L) -> int {
                    const char* path = luaL_checkstring(L, 1);
                    int res = chdir(path);
                    int last_error = (res == -1) ? errno : 0;
                    CHECK_LAST_ERROR(L, last_error, "chdir");
                    lua_pushinteger(L, res);
                    lua_pushinteger(L, last_error);
                    return 2;
                });
                return 1;
            }},
      {"ENFILE",  EMILUA_DETAIL_INT_CONSTANT(ENFILE)},
      {"MOUNT_ATTR_RDONLY",  EMILUA_DETAIL_INT_CONSTANT(MOUNT_ATTR_RDONLY)},
      {"",{}}, {"",{}},
      {"CLONE_NEWCGROUP",  EMILUA_DETAIL_INT_CONSTANT(CLONE_NEWCGROUP)},
      {"",{}},
      {"MOUNT_ATTR_STRICTATIME", 
            EMILUA_DETAIL_INT_CONSTANT(MOUNT_ATTR_STRICTATIME)},
      {"",{}}, {"",{}},
      {"EADDRINUSE",  EMILUA_DETAIL_INT_CONSTANT(EADDRINUSE)},
      {"umount", 
            [](lua_State* L) -> int {
                lua_pushcfunction(L, [](lua_State* L) -> int {
                    const char* target = luaL_checkstring(L, 1);
                    int res = umount(target);
                    int last_error = (res == -1) ? errno : 0;
                    CHECK_LAST_ERROR(L, last_error, "umount");
                    lua_pushinteger(L, res);
                    lua_pushinteger(L, last_error);
                    return 2;
                });
                return 1;
            }},
      {"EISCONN",  EMILUA_DETAIL_INT_CONSTANT(EISCONN)},
      {"",{}}, {"",{}},
      {"ENXIO",  EMILUA_DETAIL_INT_CONSTANT(ENXIO)},
      {"O_PATH",  EMILUA_DETAIL_INT_CONSTANT(O_PATH)},
      {"fsmount", 
            [](lua_State* L) -> int {
                lua_pushcfunction(L, [](lua_State* L) -> int {
                    int fd = luaL_checkinteger(L, 1);
                    unsigned int flags = luaL_checkinteger(L, 2);
                    unsigned int msflags = luaL_checkinteger(L, 3);

                    int res = fsmount(fd, flags, msflags);
                    int last_error = (res == -1) ? errno : 0;
                    CHECK_LAST_ERROR(L, last_error, "fsmount");
                    lua_pushinteger(L, res);
                    lua_pushinteger(L, last_error);
                    return 2;
                });
                return 1;
            }},
      {"",{}}, {"",{}},
      {"EFBIG",  EMILUA_DETAIL_INT_CONSTANT(EFBIG)},
      {"ERANGE",  EMILUA_DETAIL_INT_CONSTANT(ERANGE)},
      {"S_IXOTH",  EMILUA_DETAIL_INT_CONSTANT(S_IXOTH)},
      {"ENOTCONN",  EMILUA_DETAIL_INT_CONSTANT(ENOTCONN)},
      {"",{}},
      {"chmod", 
            [](lua_State* L) -> int {
                lua_pushcfunction(L, [](lua_State* L) -> int {
                    const char* path = luaL_checkstring(L, 1);
                    mode_t mode = luaL_checkinteger(L, 2);
                    int res = chmod(path, mode);
                    int last_error = (res == -1) ? errno : 0;
                    CHECK_LAST_ERROR(L, last_error, "chmod");
                    lua_pushinteger(L, res);
                    lua_pushinteger(L, last_error);
                    return 2;
                });
                return 1;
            }},
      {"EEXIST",  EMILUA_DETAIL_INT_CONSTANT(EEXIST)},
      {"MS_MOVE",  EMILUA_DETAIL_INT_CONSTANT(MS_MOVE)},
      {"AT_EMPTY_PATH",  EMILUA_DETAIL_INT_CONSTANT(AT_EMPTY_PATH)},
      {"",{}},
      {"O_NOFOLLOW",  EMILUA_DETAIL_INT_CONSTANT(O_NOFOLLOW)},
      {"",{}},
      {"ENOEXEC",  EMILUA_DETAIL_INT_CONSTANT(ENOEXEC)},
      {"O_APPEND",  EMILUA_DETAIL_INT_CONSTANT(O_APPEND)},
      {"",{}},
      {"mknod", 
            [](lua_State* L) -> int {
                lua_pushcfunction(L, [](lua_State* L) -> int {
                    const char* path = luaL_checkstring(L, 1);
                    mode_t mode = luaL_checkinteger(L, 2);
                    dev_t dev = luaL_checkinteger(L, 3);
                    int res = mknod(path, mode, dev);
                    int last_error = (res == -1) ? errno : 0;
                    CHECK_LAST_ERROR(L, last_error, "mknod");
                    lua_pushinteger(L, res);
                    lua_pushinteger(L, last_error);
                    return 2;
                });
                return 1;
            }},
      {"MS_LAZYTIME",  EMILUA_DETAIL_INT_CONSTANT(MS_LAZYTIME)},
      {"AT_RECURSIVE",  EMILUA_DETAIL_INT_CONSTANT(AT_RECURSIVE)},
      {"EADDRNOTAVAIL",  EMILUA_DETAIL_INT_CONSTANT(EADDRNOTAVAIL)},
      {"",{}},
      {"MOVE_MOUNT_SET_GROUP", 
            EMILUA_DETAIL_INT_CONSTANT(MOVE_MOUNT_SET_GROUP)},
      {"MOVE_MOUNT_T_SYMLINKS", 
            EMILUA_DETAIL_INT_CONSTANT(MOVE_MOUNT_T_SYMLINKS)},
      {"makedev", 
            [](lua_State* L) -> int {
                lua_pushcfunction(L, [](lua_State* L) -> int {
                    int major = luaL_checkinteger(L, 1);
                    int minor = luaL_checkinteger(L, 2);

                    lua_pushinteger(L, makedev(major, minor));
                    return 1;
                });
                return 1;
            }},
      {"ENETDOWN",  EMILUA_DETAIL_INT_CONSTANT(ENETDOWN)},
      {"",{}}, {"",{}}, {"",{}},
      {"EDEADLK",  EMILUA_DETAIL_INT_CONSTANT(EDEADLK)},
      {"",{}},
      {"FSCONFIG_SET_STRING", 
            EMILUA_DETAIL_INT_CONSTANT(FSCONFIG_SET_STRING)},
      {"",{}},
      {"MOUNT_ATTR_IDMAP",  EMILUA_DETAIL_INT_CONSTANT(MOUNT_ATTR_IDMAP)},
      {"S_IFBLK",  EMILUA_DETAIL_INT_CONSTANT(S_IFBLK)},
      {"MOVE_MOUNT_T_EMPTY_PATH", 
            EMILUA_DETAIL_INT_CONSTANT(MOVE_MOUNT_T_EMPTY_PATH)},
      {"MOUNT_ATTR_RELATIME", 
            EMILUA_DETAIL_INT_CONSTANT(MOUNT_ATTR_RELATIME)},
      {"",{}},
      {"S_IWOTH",  EMILUA_DETAIL_INT_CONSTANT(S_IWOTH)},
      {"umount2", 
            [](lua_State* L) -> int {
                lua_pushcfunction(L, [](lua_State* L) -> int {
                    const char* target = luaL_checkstring(L, 1);
                    int flags = luaL_checkinteger(L, 2);
                    int res = umount2(target, flags);
                    int last_error = (res == -1) ? errno : 0;
                    CHECK_LAST_ERROR(L, last_error, "umount2");
                    lua_pushinteger(L, res);
                    lua_pushinteger(L, last_error);
                    return 2;
                });
                return 1;
            }},
      {"",{}}, {"",{}}, {"",{}},
      {"ENOLCK",  EMILUA_DETAIL_INT_CONSTANT(ENOLCK)},
      {"",{}}, {"",{}},
      {"MS_SHARED",  EMILUA_DETAIL_INT_CONSTANT(MS_SHARED)},
      {"UMOUNT_NOFOLLOW",  EMILUA_DETAIL_INT_CONSTANT(UMOUNT_NOFOLLOW)},
      {"MS_MANDLOCK",  EMILUA_DETAIL_INT_CONSTANT(MS_MANDLOCK)},
      {"",{}},
      {"EALREADY",  EMILUA_DETAIL_INT_CONSTANT(EALREADY)},
      {"FSCONFIG_CMD_RECONFIGURE", 
            EMILUA_DETAIL_INT_CONSTANT(FSCONFIG_CMD_RECONFIGURE)},
      {"ENOTRECOVERABLE",  EMILUA_DETAIL_INT_CONSTANT(ENOTRECOVERABLE)},
      {"EMFILE",  EMILUA_DETAIL_INT_CONSTANT(EMFILE)},
      {"FSCONFIG_SET_PATH",  EMILUA_DETAIL_INT_CONSTANT(FSCONFIG_SET_PATH)},
      {"MOVE_MOUNT_BENEATH", 
            EMILUA_DETAIL_INT_CONSTANT(MOVE_MOUNT_BENEATH)},
      {"EOWNERDEAD",  EMILUA_DETAIL_INT_CONSTANT(EOWNERDEAD)},
      {"",{}}, {"",{}},
      {"CLONE_NEWIPC",  EMILUA_DETAIL_INT_CONSTANT(CLONE_NEWIPC)},
      {"FSCONFIG_SET_PATH_EMPTY", 
            EMILUA_DETAIL_INT_CONSTANT(FSCONFIG_SET_PATH_EMPTY)},
      {"EOVERFLOW",  EMILUA_DETAIL_INT_CONSTANT(EOVERFLOW)},
      {"O_RDWR",  EMILUA_DETAIL_INT_CONSTANT(O_RDWR)},
      {"ENOSYS",  EMILUA_DETAIL_INT_CONSTANT(ENOSYS)},
      {"ETXTBSY",  EMILUA_DETAIL_INT_CONSTANT(ETXTBSY)},
      {"FSPICK_SYMLINK_NOFOLLOW", 
            EMILUA_DETAIL_INT_CONSTANT(FSPICK_SYMLINK_NOFOLLOW)},
      {"",{}}, {"",{}},
      {"O_SYNC",  EMILUA_DETAIL_INT_CONSTANT(O_SYNC)},
      {"",{}}, {"",{}},
      {"MS_NOEXEC",  EMILUA_DETAIL_INT_CONSTANT(MS_NOEXEC)},
      {"",{}}, {"",{}}, {"",{}},
      {"SECBIT_NOROOT",  EMILUA_DETAIL_INT_CONSTANT(SECBIT_NOROOT)},
      {"FSCONFIG_SET_BINARY", 
            EMILUA_DETAIL_INT_CONSTANT(FSCONFIG_SET_BINARY)},
      {"FSMOUNT_CLOEXEC",  EMILUA_DETAIL_INT_CONSTANT(FSMOUNT_CLOEXEC)},
      {"",{}},
      {"O_DSYNC",  EMILUA_DETAIL_INT_CONSTANT(O_DSYNC)},
      {"O_NOCTTY",  EMILUA_DETAIL_INT_CONSTANT(O_NOCTTY)},
      {"MS_SYNCHRONOUS",  EMILUA_DETAIL_INT_CONSTANT(MS_SYNCHRONOUS)},
      {"",{}},
      {"S_IRWXO",  EMILUA_DETAIL_INT_CONSTANT(S_IRWXO)},
      {"S_IXUSR",  EMILUA_DETAIL_INT_CONSTANT(S_IXUSR)},
      {"",{}},
      {"MS_PRIVATE",  EMILUA_DETAIL_INT_CONSTANT(MS_PRIVATE)},
      {"SECBIT_NOROOT_LOCKED", 
            EMILUA_DETAIL_INT_CONSTANT(SECBIT_NOROOT_LOCKED)},
      {"EINPROGRESS",  EMILUA_DETAIL_INT_CONSTANT(EINPROGRESS)},
      {"MS_SLAVE",  EMILUA_DETAIL_INT_CONSTANT(MS_SLAVE)},
      {"",{}}, {"",{}}, {"",{}}, {"",{}},
      {"ENAMETOOLONG",  EMILUA_DETAIL_INT_CONSTANT(ENAMETOOLONG)},
      {"MS_NOSYMFOLLOW",  EMILUA_DETAIL_INT_CONSTANT(MS_NOSYMFOLLOW)},
      {"",{}}, {"",{}},
      {"EINVAL",  EMILUA_DETAIL_INT_CONSTANT(EINVAL)},
      {"symlink", 
            [](lua_State* L) -> int {
                lua_pushcfunction(L, [](lua_State* L) -> int {
                    const char* target = luaL_checkstring(L, 1);
                    const char* linkpath = luaL_checkstring(L, 2);
                    int res = symlink(target, linkpath);
                    int last_error = (res == -1) ? errno : 0;
                    CHECK_LAST_ERROR(L, last_error, "symlink");
                    lua_pushinteger(L, res);
                    lua_pushinteger(L, last_error);
                    return 2;
                });
                return 1;
            }},
      {"setdomainname", 
            [](lua_State* L) -> int {
                lua_pushcfunction(L, [](lua_State* L) -> int {
                    std::size_t len;
                    const char* str = lua_tolstring(L, 1, &len);
                    int res = setdomainname(str, len);
                    int last_error = (res == -1) ? errno : 0;
                    CHECK_LAST_ERROR(L, last_error, "setdomainname");
                    lua_pushinteger(L, res);
                    lua_pushinteger(L, last_error);
                    return 2;
                });
                return 1;
            }},
      {"",{}}, {"",{}}, {"",{}}, {"",{}},
      {"MOUNT_ATTR_NOATIME", 
            EMILUA_DETAIL_INT_CONSTANT(MOUNT_ATTR_NOATIME)},
      {"link", 
            [](lua_State* L) -> int {
                lua_pushcfunction(L, [](lua_State* L) -> int {
                    const char* oldpath = luaL_checkstring(L, 1);
                    const char* newpath = luaL_checkstring(L, 2);
                    int res = link(oldpath, newpath);
                    int last_error = (res == -1) ? errno : 0;
                    CHECK_LAST_ERROR(L, last_error, "link");
                    lua_pushinteger(L, res);
                    lua_pushinteger(L, last_error);
                    return 2;
                });
                return 1;
            }},
      {"",{}}, {"",{}},
      {"O_CREAT",  EMILUA_DETAIL_INT_CONSTANT(O_CREAT)},
      {"",{}}, {"",{}},
      {"ECONNRESET",  EMILUA_DETAIL_INT_CONSTANT(ECONNRESET)},
      {"S_IXGRP",  EMILUA_DETAIL_INT_CONSTANT(S_IXGRP)},
      {"S_IFCHR",  EMILUA_DETAIL_INT_CONSTANT(S_IFCHR)},
      {"",{}}, {"",{}}, {"",{}},
      {"S_IWUSR",  EMILUA_DETAIL_INT_CONSTANT(S_IWUSR)},
      {"ECONNREFUSED",  EMILUA_DETAIL_INT_CONSTANT(ECONNREFUSED)},
      {"",{}}, {"",{}},
      {"EBADF",  EMILUA_DETAIL_INT_CONSTANT(EBADF)},
      {"EAGAIN",  EMILUA_DETAIL_INT_CONSTANT(EAGAIN)},
      {"",{}}, {"",{}}, {"",{}}, {"",{}},
      {"MOVE_MOUNT_F_SYMLINKS", 
            EMILUA_DETAIL_INT_CONSTANT(MOVE_MOUNT_F_SYMLINKS)},
      {"O_WRONLY",  EMILUA_DETAIL_INT_CONSTANT(O_WRONLY)},
      {"",{}},
      {"FSPICK_NO_AUTOMOUNT", 
            EMILUA_DETAIL_INT_CONSTANT(FSPICK_NO_AUTOMOUNT)},
      {"",{}}, {"",{}}, {"",{}},
      {"AT_FDCWD",  EMILUA_DETAIL_INT_CONSTANT(AT_FDCWD)},
      {"",{}}, {"",{}}, {"",{}}, {"",{}},
      {"MOVE_MOUNT_F_EMPTY_PATH", 
            EMILUA_DETAIL_INT_CONSTANT(MOVE_MOUNT_F_EMPTY_PATH)},
      {"",{}},
      {"FSCONFIG_SET_FD",  EMILUA_DETAIL_INT_CONSTANT(FSCONFIG_SET_FD)},
      {"",{}},
      {"FSCONFIG_SET_FLAG",  EMILUA_DETAIL_INT_CONSTANT(FSCONFIG_SET_FLAG)},
      {"",{}}, {"",{}},
      {"S_IWGRP",  EMILUA_DETAIL_INT_CONSTANT(S_IWGRP)},
      {"mkfifo", 
            [](lua_State* L) -> int {
                lua_pushcfunction(L, [](lua_State* L) -> int {
                    const char* path = luaL_checkstring(L, 1);
                    mode_t mode = luaL_checkinteger(L, 2);
                    int res = mkfifo(path, mode);
                    int last_error = (res == -1) ? errno : 0;
                    CHECK_LAST_ERROR(L, last_error, "mkfifo");
                    lua_pushinteger(L, res);
                    lua_pushinteger(L, last_error);
                    return 2;
                });
                return 1;
            }},
      {"",{}}, {"",{}}, {"",{}},
      {"EXDEV",  EMILUA_DETAIL_INT_CONSTANT(EXDEV)},
      {"O_LARGEFILE",  EMILUA_DETAIL_INT_CONSTANT(O_LARGEFILE)},
      {"cap_reset_ambient", 
            [](lua_State* L) -> int {
                lua_pushcfunction(L, [](lua_State* L) -> int {
                    int res = cap_reset_ambient();
                    int last_error = (res == -1) ? errno : 0;
                    CHECK_LAST_ERROR(L, last_error, "cap_reset_ambient");
                    lua_pushinteger(L, res);
                    lua_pushinteger(L, last_error);
                    return 2;
                });
                return 1;
            }},
      {"",{}}, {"",{}}, {"",{}}, {"",{}}, {"",{}}, {"",{}},
      {"",{}},
      {"mkdir", 
            [](lua_State* L) -> int {
                lua_pushcfunction(L, [](lua_State* L) -> int {
                    const char* path = luaL_checkstring(L, 1);
                    mode_t mode = luaL_checkinteger(L, 2);
                    int res = mkdir(path, mode);
                    int last_error = (res == -1) ? errno : 0;
                    CHECK_LAST_ERROR(L, last_error, "mkdir");
                    lua_pushinteger(L, res);
                    lua_pushinteger(L, last_error);
                    return 2;
                });
                return 1;
            }},
      {"",{}}, {"",{}},
      {"MOVE_MOUNT_T_AUTOMOUNTS", 
            EMILUA_DETAIL_INT_CONSTANT(MOVE_MOUNT_T_AUTOMOUNTS)},
      {"",{}}, {"",{}}, {"",{}}, {"",{}}, {"",{}}, {"",{}},
      {"",{}}, {"",{}},
      {"SECBIT_NO_SETUID_FIXUP", 
            EMILUA_DETAIL_INT_CONSTANT(SECBIT_NO_SETUID_FIXUP)},
      {"",{}}, {"",{}}, {"",{}}, {"",{}},
      {"ECONNABORTED",  EMILUA_DETAIL_INT_CONSTANT(ECONNABORTED)},
      {"",{}},
      {"SECBIT_NO_SETUID_FIXUP_LOCKED", 
            EMILUA_DETAIL_INT_CONSTANT(SECBIT_NO_SETUID_FIXUP_LOCKED)},
      {"",{}}, {"",{}}, {"",{}}, {"",{}}, {"",{}},
      {"EBUSY",  EMILUA_DETAIL_INT_CONSTANT(EBUSY)},
      {"O_EXCL",  EMILUA_DETAIL_INT_CONSTANT(O_EXCL)},
      {"",{}}, {"",{}}, {"",{}},
      {"umask", 
            [](lua_State* L) -> int {
                lua_pushcfunction(L, [](lua_State* L) -> int {
                    mode_t res = umask(luaL_checkinteger(L, 1));
                    lua_pushinteger(L, res);
                    return 1;
                });
                return 1;
            }},
      {"",{}}, {"",{}}, {"",{}}, {"",{}}, {"",{}},
      {"S_IRWXU",  EMILUA_DETAIL_INT_CONSTANT(S_IRWXU)},
      {"",{}}, {"",{}},
      {"FSCONFIG_CMD_CREATE", 
            EMILUA_DETAIL_INT_CONSTANT(FSCONFIG_CMD_CREATE)},
      {"",{}},
      {"EFAULT",  EMILUA_DETAIL_INT_CONSTANT(EFAULT)},
      {"",{}}, {"",{}},
      {"FSCONFIG_CMD_CREATE_EXCL", 
            EMILUA_DETAIL_INT_CONSTANT(FSCONFIG_CMD_CREATE_EXCL)},
      {"",{}}, {"",{}},
      {"S_ISVTX",  EMILUA_DETAIL_INT_CONSTANT(S_ISVTX)},
      {"",{}}, {"",{}}, {"",{}}, {"",{}},
      {"O_ASYNC",  EMILUA_DETAIL_INT_CONSTANT(O_ASYNC)},
      {"",{}}, {"",{}}, {"",{}}, {"",{}}, {"",{}}, {"",{}},
      {"",{}}, {"",{}},
      {"SECBIT_KEEP_CAPS",  EMILUA_DETAIL_INT_CONSTANT(SECBIT_KEEP_CAPS)},
      {"",{}}, {"",{}}, {"",{}}, {"",{}}, {"",{}},
      {"EAFNOSUPPORT",  EMILUA_DETAIL_INT_CONSTANT(EAFNOSUPPORT)},
      {"SECBIT_KEEP_CAPS_LOCKED", 
            EMILUA_DETAIL_INT_CONSTANT(SECBIT_KEEP_CAPS_LOCKED)},
      {"",{}},
      {"ESRCH",  EMILUA_DETAIL_INT_CONSTANT(ESRCH)},
      {"",{}},
      {"SECBIT_NO_CAP_AMBIENT_RAISE", 
            EMILUA_DETAIL_INT_CONSTANT(SECBIT_NO_CAP_AMBIENT_RAISE)},
      {"",{}},
      {"FSOPEN_CLOEXEC",  EMILUA_DETAIL_INT_CONSTANT(FSOPEN_CLOEXEC)},
      {"",{}}, {"",{}}, {"",{}}, {"",{}},
      {"SECBIT_NO_CAP_AMBIENT_RAISE_LOCKED", 
            EMILUA_DETAIL_INT_CONSTANT(SECBIT_NO_CAP_AMBIENT_RAISE_LOCKED)},
      {"",{}}, {"",{}}, {"",{}}, {"",{}}, {"",{}}, {"",{}},
      {"",{}}, {"",{}}, {"",{}}, {"",{}}, {"",{}}, {"",{}},
      {"",{}}, {"",{}}, {"",{}}, {"",{}}, {"",{}}, {"",{}},
      {"",{}}, {"",{}}, {"",{}},
      {"ECHILD",  EMILUA_DETAIL_INT_CONSTANT(ECHILD)},
      {"",{}}, {"",{}}, {"",{}}, {"",{}}, {"",{}}, {"",{}},
      {"MOVE_MOUNT_F_AUTOMOUNTS", 
            EMILUA_DETAIL_INT_CONSTANT(MOVE_MOUNT_F_AUTOMOUNTS)},
      {"",{}}, {"",{}}, {"",{}},
      {"EHOSTUNREACH",  EMILUA_DETAIL_INT_CONSTANT(EHOSTUNREACH)},
      {"",{}},
      {"AT_SYMLINK_NOFOLLOW", 
            EMILUA_DETAIL_INT_CONSTANT(AT_SYMLINK_NOFOLLOW)},
      {"",{}},
      {"S_IRWXG",  EMILUA_DETAIL_INT_CONSTANT(S_IRWXG)},
      {"",{}}, {"",{}}, {"",{}}, {"",{}}, {"",{}}, {"",{}},
      {"",{}}, {"",{}}, {"",{}}, {"",{}}, {"",{}}, {"",{}},
      {"",{}}, {"",{}}, {"",{}}, {"",{}}, {"",{}},
      {"O_CLOEXEC",  EMILUA_DETAIL_INT_CONSTANT(O_CLOEXEC)},
      {"",{}}, {"",{}},
      {"EBADMSG",  EMILUA_DETAIL_INT_CONSTANT(EBADMSG)},
      {"",{}}, {"",{}}, {"",{}}, {"",{}}, {"",{}}, {"",{}},
      {"",{}}, {"",{}}, {"",{}}, {"",{}}, {"",{}}, {"",{}},
      {"",{}}, {"",{}}, {"",{}}, {"",{}}, {"",{}}, {"",{}},
      {"EACCES",  EMILUA_DETAIL_INT_CONSTANT(EACCES)},
      {"FSPICK_EMPTY_PATH",  EMILUA_DETAIL_INT_CONSTANT(FSPICK_EMPTY_PATH)},
      {"",{}}, {"",{}}, {"",{}}, {"",{}}, {"",{}}, {"",{}},
      {"",{}}, {"",{}}, {"",{}}, {"",{}}, {"",{}}, {"",{}},
      {"",{}}, {"",{}}, {"",{}}, {"",{}}, {"",{}}, {"",{}},
      {"",{}}, {"",{}}, {"",{}}, {"",{}},
      {"EWOULDBLOCK",  EMILUA_DETAIL_INT_CONSTANT(EWOULDBLOCK)},
      {"",{}}, {"",{}}, {"",{}}, {"",{}}, {"",{}}, {"",{}},
      {"",{}}, {"",{}}, {"",{}}, {"",{}}, {"",{}}, {"",{}},
      {"",{}}, {"",{}}, {"",{}}, {"",{}}, {"",{}}, {"",{}},
      {"",{}}, {"",{}}, {"",{}}, {"",{}}, {"",{}}, {"",{}},
      {"",{}}, {"",{}}, {"",{}}, {"",{}}, {"",{}}, {"",{}},
      {"",{}}, {"",{}}, {"",{}}, {"",{}}, {"",{}}, {"",{}},
      {"",{}}, {"",{}},
      {"ECANCELED",  EMILUA_DETAIL_INT_CONSTANT(ECANCELED)},
      {"",{}}, {"",{}}, {"",{}}, {"",{}}, {"",{}}, {"",{}},
      {"",{}}, {"",{}}, {"",{}}, {"",{}}, {"",{}}, {"",{}},
      {"",{}}, {"",{}}, {"",{}}, {"",{}}, {"",{}}, {"",{}},
      {"",{}},
      {"FSPICK_CLOEXEC",  EMILUA_DETAIL_INT_CONSTANT(FSPICK_CLOEXEC)}
    };

  if (len <= MAX_WORD_LENGTH && len >= MIN_WORD_LENGTH)
    {
      unsigned int key = hash (str, len);

      if (key <= MAX_HASH_VALUE)
        {
          const char *s = wordlist[key].name;

          if (*str == *s && !strcmp (str + 1, s + 1))
            return &wordlist[key];
        }
    }
  return 0;
}

} // namespace emilua::gperf::detail


namespace emilua {





int posix_mt_index(lua_State* L)
{
    auto key = tostringview(L, 2);
    return  ::emilua::gperf::detail::value_or(::emilua::gperf::detail::gperf__src_linux_ipc_actor_sys_bindings_cpp_1::in_word_set((key).data(), (key).size()), [](lua_State* L) -> int {
            return luaL_error(L, "key not found");
        })(L);
}

} // namespace emilua
